local MAJOR = "StarLibWidgetIcon-1.0"
local MINOR = 14

assert(LibStub, MAJOR.." requires LibStub")
local WidgetIcon = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetIcon then return end
local LibWidget = LibStub("StarLibWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires StarLibWidget-1.0")
local LibProperty = LibStub("StarLibProperty-1.0", true)
assert(LibProperty, MAJOR .. " requires StarLibProperty-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local Evaluator = LibStub("StarLibEvaluator-1.0", true)
assert(Evaluator, MAJOR .. " requires StarLibEvaluator-1.0")

local pool = setmetatable({}, {__mode = "k"})

if not WidgetIcon.__index then
	WidgetIcon.__index = WidgetIcon
end

WidgetIcon.defaults = {
	speed = 500,
	repeating = true
}

local widgetType = {icon=true}

local function Split(str, delim)
	local start = 1
	local tbl = {}
	for i = 2, strlen(str) do
		if str:sub(i, i) == delim then
			tinsert(tbl, str:sub(start, i - 1))
			start = i + 1
		end
	end

	if start < strlen(str) then
		tinsert(tbl, str:sub(start, strlen(str)))
	end

	return tbl
end

-- Create a new StarLibWidgetIcon object
-- @name WidgetIcon.New
-- @usage WidgetIcon:New(visitor, name, config, row, col, layer, errorLevel)
-- @param visitor An StarLibCore-1.0 object
-- @param config This widget's settings
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param self.errorLevel The self.errorLevel for this object
-- @param callback Your draw function. The widget is passed as first parameter and data is passed as 2nd.
-- @param data Some data to pass to the draw function
-- @return A new StarLibWidgetIcon object

function WidgetIcon:New(visitor, name, config, row, col, layer, errorLevel, callback, data, timer)
	assert(name, "WidgetIcon requires a name.")
	assert(config, "Please provide the icon with a config")

	local obj = next(pool)

	if obj then
		pool[obj] = nil
		obj.__index = nil
	else
		obj = {}
		obj.options = {}
	end

	setmetatable(obj, self)

	obj.widget = LibWidget:New(obj, visitor, name, config, row, col, layer, widgetType, errorLevel)

	obj.config = config
	obj.callback = callback
	obj.data = data

	obj:Init()

	return obj
end

function WidgetIcon:Init()

	self.cols = 1;
	self.rows = 1;

	self.speed = self.config.speed or self.defaults.speed

	self.fg_color = self.config.foreground or WidgetIcon.defaults.background
	self.bg_color = self.config.background or WidgetIcon.defaults.background

	local val = self.config.bitmap

	if( not val) then
		update_ = -1;
		return;
	end

	self.bitmap = val

	local ch = {}

	for i = 1, self.visitor.lcd.YRES do
		local row = val["row" .. i]
		if not row then break end
		local line = Split(row, "|")
		for j, segment in ipairs(line) do
			if j > #ch then
				tinsert(ch, {}) --SpecialChar:New(visitor.lcd.YRES))
				for i = 1, self.visitor.lcd.YRES do
					ch[j][i] = 0
				end
			end
			for c = 1, strlen(segment) do
				if segment:sub(c, c) == "*" then
					ch[j][i] = bit.bxor(ch[j][i], bit.lshift(1, c - 1))
				end
			end
		end
	end

	self.icon = ch

	self.index = 1
	self.started = false
	self.ch = -1
	self.repeating = self.config.repeating or self.defaults.repeating
	self.speed = self.config.speed or self.defaults.speed

	self.bitmap = {} --SpecialChar(self.data[0]:Size())

	self.timer = self.config.timer or LibTimer:New("WidgetText.timer " .. self.name, self.speed, self.repeating, self.Update, self, self.errorLevel)

	self.visible = LibProperty:New(self, self.visitor, self.name, self.config.visible, "return true", self.errorLevel);

	--QObject::connect(visitor_->GetWrapper(), SIGNAL(_ResizeLCD(int, int, int, int)),
	--	this, SLOT(Resize(int, int, int, int)));

	return obj
end



-- Delete a StarLibWidgetIcon object
-- @name StarLibWidgetIcon.Del
-- @usage StarLibWidgetIcon:Del([marq])
-- @param marq A StarLibWidgetIcon object
function WidgetIcon:Del(key)
	if not key then
		key = self
	end
	key.timer:Del()
	key.visible:Del()
	key.widget:Del()
	key:Stop()
	pool[key] = true
end

-- Start a StarLibWidgetIcon
-- @name StarLibWidgetIcon.Start
-- @usage object:Start()
-- @return Nothing
function WidgetIcon:Start()
	self.error:Print("WidgetIcon:Start")
	if(self.speed <= 0 or self.active) then
		return;
	end
	self.timer:Start();
	self.enabled = true;
	self.active = true
	self:Update();
end

-- Stop a StarLibWidgetIcon
-- @name StarLibWidgetIcon.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetIcon:Stop()
	self.error:Print("WidgetIcon:Stop")
	self.timer:Stop()
	self.enabled = false
	self.active = false
end

function WidgetIcon:Resize(rows, cols, old_rows, old_cols)
	local yres = self.visitor.YRES;
	local xres = self.visitor.XRES;
	local r = self.row * yres / old_rows;
	local c = self.col * xres / old_cols;
	self.row = round(rows * r / yres);
	self.col = round(cols * c / xres);
	self:Update();
end

function WidgetIcon:SetupChars()
	--[[for i, widget in ipairs(self.visitor.widgets) do
		if widget.widget_base == self.widget_base then
			self.start = widget.start
			self.finish = widget.finish
			return
		end
	end]]
	self.start = #self.visitor.lcd.specialChars
	for i, v in ipairs(self.icon) do
		tinsert(self.visitor.lcd.specialChars, v)
	end
	self.finish = #self.visitor.lcd.specialChars
end

function WidgetIcon:Draw()
	if type(self.callback) == "function" then
		self.callback(self, self.data)
	end
end

function WidgetIcon:Update()
	if not self.active then return end

	self:Draw();

	self.index = self.index + 1
	if(self.index > #self.icon) then
		self.index = 1;
	end
end

function WidgetIcon:GetOptions(visitor, db)
	local defaults = WidgetIcon.defaults
	local options = {
		enabled = {
			name = "Enabled",
			desc = "Whether this icon is enabled or not",
			type = "toggle",
			get = function() return db.enabled end,
			set = function(info, v) db.enabled = v; db["enabledDirty"] = true end,
			order = 1
		}
	}
	options.widget = LibWidget:GetOptions(visitor, db)
	return options
end

